#ifndef _hfe_h
#define _hfe_h

#include "osihfe.h"
#ifndef _M_IX86
#define _M_IX86  1 //todo:fix for big_endian linux
#endif

#ifdef  _MSC_VER
/*
 * Currently, all MS C compilers for Win32 platforms default to 8 byte
 * alignment.
 */
#pragma pack(push,1)
#endif  /* _MSC_VER */

#if defined(__BYTE_ORDER) && __BYTE_ORDER == __LITTLE_ENDIAN || defined (_M_IX86) || defined(__LITTLE_ENDIAN__) || defined(_WIN32) || defined( FLASH)
#define NATIVE_BYTEORDER_ENCODING 
#else
/* Probably a SUN or an old MAC */
#define REVERSE_BYTEORDER_ENCODING
void swapByteOrder(unsigned short& us)
{
    us = (us >> 8) |
         (us << 8);
}

#endif


// First part : 0x0000-0x0200 (512 bytes) : File header
// Data is stored LITTLE ENDIAN
typedef struct picfileformatheader_
{
unsigned char HEADERSIGNATURE[8]; // "HXCPICFE"
unsigned char formatrevision;	// Revision 0
unsigned char number_of_track;	// Number of track in the file
unsigned char number_of_side;	// Number of valid side (Not used by the emulator)
unsigned char track_encoding;	// Track Encoding mode
								// (Used for the write support - Please see the list above)
unsigned short bitRate;			// Bitrate in Kbit/s. Ex : 250=250000bits/s
								// Max value : 500
unsigned short floppyRPM;		// Rotation per minute (Not used by the emulator)
unsigned char floppyinterfacemode; // Floppy interface mode. (Please see the list above.)
unsigned char dnu;				// Free
unsigned short track_list_offset; // Offset of the track list LUT in block of 512 bytes
								// (Ex: 1=0x200)
unsigned char write_allowed;	// The Floppy image is write protected ?
unsigned char single_step;		// 0xFF : Single Step - 0x00 Double Step mode
unsigned char track0s0_altencoding; // 0x00 : Use an alternate track_encoding for track 0 Side 0
unsigned char track0s0_encoding; // alternate track_encoding for track 0 Side 0
unsigned char track0s1_altencoding; // 0x00 : Use an alternate track_encoding for track 0 Side 1
unsigned char track0s1_encoding; // alternate track_encoding for track 0 Side 1
}
#ifdef __APPLE__
__attribute__((packed))
#endif
picfileformatheader;

//floppyinterfacemode values :
#define IBMPC_DD_FLOPPYMODE 0x00
#define IBMPC_HD_FLOPPYMODE 0x01
#define ATARIST_DD_FLOPPYMODE 0x02
#define ATARIST_HD_FLOPPYMODE 0x03
#define AMIGA_DD_FLOPPYMODE 0x04
#define AMIGA_HD_FLOPPYMODE 0x05
#define CPC_DD_FLOPPYMODE 0x06
#define GENERIC_SHUGGART_DD_FLOPPYMODE 0x07
#define IBMPC_ED_FLOPPYMODE 0x08
#define MSX2_DD_FLOPPYMODE 0x09
#define C64_DD_FLOPPYMODE 0x0A
#define EMU_SHUGART_FLOPPYMODE 0x0B
#define S950_DD_FLOPPYMODE 0x0C
#define S950_HD_FLOPPYMODE 0x0D
#define DISABLE_FLOPPYMODE 0xFE

//track_encoding / track0s0_encoding / track0s1_encoding values :
#define ISOIBM_MFM_ENCODING 0x00
#define AMIGA_MFM_ENCODING 0x01
#define ISOIBM_FM_ENCODING 0x02
#define EMU_FM_ENCODING 0x03
#define UNKNOWN_ENCODING 0xFF

//Note :
//If track0s0_altencoding is set to 0xFF, track0s0_encoding is ignored and 
// track_encoding is used for track 0 side 0.
//If track0s1_altencoding is set to 0xFF, track0s1_encoding is ignored and 
// track_encoding is used for track 0 side 1.

//Second part : (up to 1024 bytes) : Track offset LUT
typedef struct pictrack_
{
	unsigned short offset; // Offset of the track data in block of 512 bytes (Ex: 2=0x400)
	unsigned short track_len; // Length of the track data in byte.
}
#ifdef __APPLE__
__attribute__((packed))
#endif
pictrack;
//For a disk of 80 tracks there are a table of 80 pictrack structure.
//pictrack[80];

//Third part : Track data
//A track data is a table containing the bit stream of a track of the floppy disk. A track can
//contain a MFM / FM / GCR or a custom encoding.
//The track is divided in block of 512bytes and each block contains a part of the Side 0
//track and a part of the Side 1 track:
//Figure 1 : A track data
//
// 0x0000  0x0100  0x0200  0x0300  0x0400  0x0500  0x0600  0x0700  0x0800
// | side0 | side1 | side0 | side1 | side0 | side1 | side0 | side1 |
// | TRK BLK 0     | TRK BLK 1     | TRK BLK 2     | TRK BLK 3     |
// |  512 bytes    |  512 bytes	   |  512 bytes	   |  512 bytes    |
//
//The bits transmitting order to the FDC is :
//Bit 0-> Bit 1-> Bit 2-> Bit 3-> Bit 4-> Bit 5-> Bit 6-> Bit 7->(next byte)




//OSI bitstream is programable 8E1 - 1 start bit (logic 0), 8 data, 1 parity, 1 stop bit (logic 1) (11 bits total)
//or 8N1 for OS65U 1 start bit (logic 0), 8 data bits, 1 stop bit(logic 1) (10 bits total)
//8 data bits are sent lsb to msb  (bit0, bit1, bit2, bit3, bit4, bit5, bit6, bit7)
// A clock bit is always sent before a data bit
/*
---|  |--|--|--|--|--|--|--|--|--|--|...
   |__|__|__|__|__|__|__|__|__|__| S|
								   t
	S   0  1  2  3  4  5  6  7 (P) o
 Start                             p
*/

#ifdef  _MSC_VER
#pragma pack(pop)
#endif  /* _MSC_VER */

/// WinOSI defines
typedef enum tag_DISKTYPE
{
	DISK_ERROR = -1,
	DISK_UNKNOWN = 0,
	OS65D5 = 5,
	HEXDOS = 6,
	DOS655 = 7,
	OS65D8 = 8,
	OS65U8 = 9,
	DOS658 = 10,
	EMPTY8 = 11,
	EMPTY5 = 12,
	EMPTY8Z = 13,
	EMPTY5Z = 14,
	MAX_TYPE = 15 //must be last
} DISKTYPE;



#define MAXTRK5		39		// maximum track number (0 to 39 = 40 tracks)
#define TRKSIZ5		0x0900	// image track size (less for a real disk)
#define DIRTRACK5	12		// OS65D index of directory info track
#define SECTORS5	8		// sectors per track 5.25

#define MAXTRK8		76		// maximum tack number (0 to 76 = 77 tracks)
#define TRKSIZ8		0x0f00	// image track size (less for real disk)
#define TRKSIZ8E	0x0e00	// old winosi 8" track size (Do Not Use)

#define DIRTRACK8	8		// index of directory info track
#define SECTORS8	12		// sectors per track 8"
#define DIRTRACK65U	7		// OS65D directory track 8"

#define DOS65DIRTRACK5  4  //DOS65 5.25" directory track
#define DOS65DIRTRACK8  3  //DOS65 8" directory track

#define ERR_NONE  0
#define ERR_FRAME 1
#define ERR_PARITY 2

#define DISK8INDEX 800		  //in flux nibbles (clk+0)
#define DISK5INDEX 1478		  //in flux nibbles

#define DISK5TARGETLEN 13000  //in storage bytes, 1 side 1  
//250Kbps * 2bit per flux * disk revolution time (.200ms) = 100000 /8bits per byte=12500 * 2 sides =  25000 bytes per track
//250Kbps * 1bit per flux * .2ms = 50000 / 8bits =  6250 * 2 sides = 12500 

#define DISK8TARGETLEN 20898  //in storage bytes, not flux nibbles
//500Kbps * 2bit per flux * disk revolution time (.166s) = 166000 /8bits per byte =20750 bytes	* 2 sides =  41500 bytes per track
//500Kbps * 1 bit per flux *.166ms = 83000 /8bits =  10375 bytes * 2 sides = 20750

extern const char szHeaderSig[];

//supported software ACIA encoding types
enum eSERFORMAT {
	s8E1 = 0,
	s8N1 = 1,
};

// FM decoding state for converting from HFE
typedef struct FMstate_
{
	eSERFORMAT eSformat; //serial word format
	int sWordIndex;		 //serial word bit index, 0 = lsb
	bool bSerWord[11];	//0...10 for 8E1
	bool bClock;		//flag we have FM clock pulse 
	bool bPastIndex;
	unsigned char data;	//current input character
	unsigned char error; //serial data error condition
	unsigned int nPos;	//bit index
	int nSampleCount;	//counting the source bits required for a data bit
	int nZeroRun;
	int nulRun;
	

}FMstate;

//os65d 8" = 64 entries max
//os65d 5.25 = 64 entries max
//os65u 8" = 63 files on floppy

typedef struct OSIDirMap_
{
	unsigned char Name[120][32];  //fixed length, usually 6 or 15 chars
	int Start[120]; //trk# or address offset(65u)
	int End[120];   //trk# or length(65u)
	int Errors[120]; //# errors for this entry (track data problems)
	unsigned char BAM128[0xf00 * 80 / 128]; //Block map for CP/M, DOS/65 -- 80 trks (2400 bytes) (& others)
}OSIDirMap;

void FMdecodeInitialize(FMstate& fmstate);
int FMdecodestream(const unsigned char *inBuf, int nInLen, int bitwidth, unsigned char* outBuf, unsigned char *errBuf, 
				   int nOutLen, FMstate& fmstate, int* pnRead, int *pnWrote, int pos, bool bVerbose);

DISKTYPE DetermineOSIDiskTypeFromHFE(unsigned char * pHFEDisk, int nHFELen, int nSide =0, int nSkip=1);
DISKTYPE IdentifyDiskType( unsigned char *pMem, int nReadBytes);
bool DisplayHFEFileInfo(unsigned char * pHFEDisk, int nHFELen);

int OSISeekTrackStart( const unsigned char* pTrack, int nLen, DISKTYPE dt = DISK_UNKNOWN, int iTrack = -1);

bool CreateHFEImage( unsigned char * pDisk, DISKTYPE dt, int nSides ,  int nBitwidth, unsigned char** ppData, int* pnLen, bool bWrite8As5 );
bool CreateOSIImage( unsigned char * pHFEDisk, int nHFELen, int nSide, int nSkip, unsigned char** ppOSI, int* pnOSILen, DISKTYPE* dt, bool bVerbose = false);
bool CleanOSIImage( unsigned char * pDisk, DISKTYPE dt, unsigned char** ppData, int* pnLen );
int TestOSIImage( unsigned char * pMem, DISKTYPE dt, bool bErrorTrackPresent );

TCHAR * GetOS65DDir(unsigned char * pTrack, DISKTYPE dt, OSIDirMap* pDirMap = NULL);
TCHAR * GetOS65UDir(unsigned char * pTrack, DISKTYPE dt, OSIDirMap* pDirMap = NULL);
TCHAR * GetHEXDOSDir(unsigned char * pTrack, DISKTYPE d, OSIDirMap* pDirMap = NULL);
TCHAR * GetDOS65Dir(unsigned char * pTrack, DISKTYPE dt, OSIDirMap* pDirMap = NULL);
TCHAR * GetOSIDirectory(unsigned char * pMem, DISKTYPE dt, OSIDirMap* pDirMap = NULL);


#endif // _hfe_h
