// OSIHFE.cpp : Defines the entry point for the console application.
// Mark Spankus 2018
//
#include "osihfe.h"

#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdarg.h>
#include "hfe.h"
#include "scp.h"

bool IsHexFile(unsigned char * pMem, int nLen);
int ReadHexFile( FILE* inFile, unsigned char **ppMem, int* pnLen);



FILE* logfile = stderr;
bool bVerbose = false;

const TCHAR * szIdentity[MAX_TYPE] = {
	_T("Unknown\0") ,
	_T("\0"),
	_T("\0"),
	_T("\0"),
	_T("\0"),
	_T("OS65D5\0"),
	_T("HEXDOS\0"),
	_T("DOS/65-5\0"),
	_T("OS65D8\0"),
	_T("OS65U\0"),
	_T("DOS/65-8\0"),
	_T("Empty 8\"\0"),
	_T("Empty 5.25\"\0"),
	_T("Empty 8\"\0"),
	_T("Empty 5.25\"\0"),
};


void usage()
{
	_fputts( _T("Usage: OSIHFE -[5bcdilotv] infile [outfile]\r\n       Converts between OSI disk dump and HxC floppy emulator file formats\r\n"), stdout);
	_fputts( _T("       will also read SuperCopyPro (SCP) and HEX Dump OSI disk images.\r\n"), stdout);
	_fputts( _T("       If [outfile] is not specified, a target file is automatically\r\n")\
	         _T("          created based on input filetype\r\n"), stdout);
	_fputts( _T("       -5 :write 8\" image at 5.25\" bit rate to allow mounting 8\" image on 5.25\"\r\n           system\r\n"), stdout);
	//_fputts( _T("       -8 :write 5.25\" image at 8\" bit rate to allow mounting 5.25\" image on 8\"\r\n           system\r\n"), stdout);
	_fputts( _T("       -b65D5/-b65D8/-b65U/-bHEXDOS -b8/-b8z -b5/-b5z outfile :create blank OSI\r\n           HFE image of specified type or empty (-b8/-b5) or null (-b5z/-b8z)\r\n"), stdout);
	_fputts( _T("       -c infile outfile :clean OSI disk dump image (not HFE)\r\n"), stdout);
	_fputts( _T("       -d infile :list directory contents\r\n"), stdout);
	_fputts( _T("       -i infile :identify type (OS65D5, OS65D8, OS65U, HEXDOS)\r\n"), stdout);
	_fputts( _T("       -k<n> :skip n tracks in scp source image (n=4: 160trk->40trk)\r\n"), stdout);
	_fputts( _T("       -l<filename> :specify verbose log output file name instead of stderr\r\n"), stdout);
	_fputts( _T("       -o :overwrite existing output file\r\n"), stdout);
	_fputts( _T("       -r<n> :specify source revolution to use in conversion from scp file (usually 1-3)\r\n"), stdout);
	_fputts( _T("       -s :single bit for 250K flux per second vs std 500K fps HFE\r\n"), stdout);  //support HFE 250Kbps bitrate vs std/old 500Kbps flux rate


	_fputts( _T("       -t infile :test OSI disk dump/HFE/SCP image, report errors\r\n"), stdout);
	_fputts( _T("       -v :turn on verbose output\r\n"), stdout);
	_fputts( _T(" (ver 2.21)\r\n"), stdout);

	//_fputts( _T("       -a  diskfile datafile [t=n s=n][name=\"FILNAM\"] :add file to disk image (no data manipulation)\r\n"), stdout);
	//_fputts( _T("       -ab diskfile datafile [t=n s=n][name=\"FILNAM\"] :add BASIC text file to disk image (BASIC is tokenized)\r\n"), stdout);

	//_fputts( _T("       -x  diskfile [t=n s=n][name=\"FILNAM\"] outfile :extract data from disk image by name or track\r\n"), stdout);
	//_fputts( _T("       -xb diskfile [t=n s=n][name=\"FILNAM\"] outfile :extract detokenized BASIC from disk image by name or track\r\n"), stdout);

}


int _tmain(int argc, TCHAR* argv[])
{
	
	FILE* inpfile = NULL;
	FILE* outfile = NULL;
	

	bool bBlank = false;
	bool bClean = false;
	bool bTest = false;
	bool bOverwrite = false;
	bool bIdentify = false;
	bool bWrite8As5 = false;
	bool bWrite5As8 = false;
	bool bDirectory = false;
	bool bIsSCP = false;
	bool bIsHexFile = false;
	bool bMakeHFE = false;
	bool bMake65D = false;
	int  nSkipTrack = 1;
	int  nSCPRev = 1;
	
	bool bAdd = false;
	bool bExtract = false;
	//int nTrack = -1;
	//int nSector = -1;
	//TCHAR DISKFILE[32];


	DISKTYPE dTarget = DISK_UNKNOWN;


	unsigned char* pMem = NULL;
	unsigned char *pEnd = NULL;

	int nSide = 1;
	int nBitwidth = 2;
	DISKTYPE dt = DISK_UNKNOWN;
	unsigned char * pOSIDisk = NULL;
	int nOSILen = 0;
	int nReturn = 0;
	
	TCHAR szFile[8192] = {0};
	TCHAR szOutFile[8192] = {0};


    while (--argc && ((*++argv)[0] == _T('-') )) 
	{
		const TCHAR * arg = argv[0];
		arg++;
		if (!_tcsicmp(arg, _T("v"))) 
		{          
            bVerbose = true;
			continue;
        }

		if (!_tcsicmp(arg, _T("o")))
		{
			bOverwrite = true;
			continue;
		}
		if (!_tcsicmp(arg, _T("i")))
		{
			bIdentify = true;
			continue;
		}
		if (!_tcsicmp(arg, _T("c")))
		{
			bClean = true;
			continue;
		}
		
		if (!_tcsicmp(arg, _T("t")))
		{
			bTest = true;
			continue;
		}
		if (!_tcsicmp(arg, _T("d")))
		{
			bDirectory = true;
			continue;
		}
		if (!_tcsicmp(arg, _T("s")))
		{
			nBitwidth = 1;
			continue;
		}

		if (!_tcsicmp(arg, _T("5")))
		{
			bWrite8As5 = true;
			continue;
		}
		if (!_tcsicmp(arg, _T("8")))
		{
			bWrite5As8 = true;
			continue;
		}

		if (!_tcsicmp(arg, _T("b65D5")))
		{
			bBlank = true;
			dTarget = OS65D5;
			continue;
		}
		if (!_tcsicmp(arg, _T("b65D8")))
		{
			bBlank = true;
			dTarget = OS65D8;
			continue;
		}
		if (!_tcsicmp(arg, _T("b65U")))
		{
			bBlank = true;
			dTarget= OS65U8;
			continue;
		}
		if (!_tcsicmp(arg, _T("bHEXDOS")))
		{
			bBlank = true;
			dTarget= HEXDOS;
			continue;
		}
		if (!_tcsicmp(arg, _T("b8")))
		{
			bBlank = true;
			dTarget= EMPTY8;
			continue;
		}
		if (!_tcsicmp(arg, _T("b5")))
		{
			bBlank = true;
			dTarget= EMPTY5;
			continue;
		}
		if (!_tcsicmp(arg, _T("b8z")))
		{
			bBlank = true;
			dTarget= EMPTY8Z;
			continue;
		}
		if (!_tcsicmp(arg, _T("b5z")))
		{
			bBlank = true;
			dTarget= EMPTY5Z;
			continue;
		}

		if (!_tcsicmp(arg, _T("a")))
		{
			bAdd = true;
			continue;
		}
		if (!_tcsicmp(arg, _T("x")))
		{
			bExtract = true;
			continue;
		}
		if (arg[0] == _T('k') || arg[0] == _T('K'))
		{
			if ( arg[1] && arg[1] >= _T('0') && arg[1] <= _T('9'))
				nSkipTrack = (int) arg[1] - _T('0');
			continue;
		}
		 
		if (arg[0] == _T('r') || arg[0] == _T('R'))
		{
			if ( arg[1] && arg[1] >= _T('0') && arg[1] <= _T('9'))
				nSCPRev = (int) arg[1] - _T('0');
			continue;
		}

		
		if ( (*arg == _T('l') || *arg == _T('L'))  && _tcslen(arg) > 1) //logfile
		{
			if ( (logfile = _tfopen(arg+1, _T("wb"))) == NULL )
				logfile = stderr;
			continue;

		}

		if (bTest && bClean)
		{
			_fputts( _T(" Only one Test/Clean op allowed.\r\n"), stderr);
			return 0;
	
		}
		if (bWrite5As8 && bWrite8As5)
		{
			_fputts( _T(" Only one forced bitrate op allowed. (-5 or -8)\r\n"), stderr);
			return 0;

		}

		_fputts( _T(" Unknown argument:"), stderr);
		_fputts( arg, stderr);
		_fputts( _T("\r\n"), stderr);

		//anything else
		usage();
		return 0;


	}	
	
	if (argc > 0)
	{
	    int i = 0;

		_tcsncpy(szFile, argv[i], 8191);

		if (!bBlank)
		{
			if ( (inpfile = _tfopen(argv[i], _T("rb"))) == NULL)
			{
				_ftprintf(stderr, _T("osihfe: Can't open '%s' for input.\r\n"), argv[i]);
				return -1;
			}
		}
		else
		{
			if (argc != 1)
			{
				_ftprintf(stderr, _T("osihfe: Too many file arguments for operation.\r\n"));	
				return -1;
			}
			i--;
		}


		if ( ++i < argc && !bTest)
		{
			
			_tcsncpy(szOutFile, argv[i], 8191);
			//see if we can determine target format based on output file type
			//for not 65d, hfe input files (SCP, HEX)
			int nLen = _tcslen(szOutFile);
			int j = nLen-1;
			while (	j > 0 && szOutFile[j] != _T('.'))
				j--;
			if (j)
			{
				if (!_tcsicmp(szOutFile, _T(".scp")))
				{
					_ftprintf(stderr, _T("osihfe: SCP as target file format not supported.\r\n"));
					return -1;
				}
				if (!_tcsicmp(szOutFile, _T(".hfe")))
				{
					bMakeHFE = true;
				}
				else if (!_tcsicmp(szOutFile, _T(".65d")) ||
						 !_tcsicmp(szOutFile, _T(".65u")) )
				{
					bMake65D = true;
				} 
			}

			if ( (outfile = _tfopen(argv[i], _T("r+b"))) != NULL )
			{
				if (!bOverwrite)
				{
					if (outfile)
					{
						fclose(outfile);
						outfile = NULL;
					}
					if (inpfile)
					{
						fclose(inpfile);
						inpfile = NULL;
					}
					_ftprintf(stderr, _T("osihfe: Target file '%s' exists. Use -o to overwrite\r\n"), argv[i]);
					return -1;
				}
				//close file to reopen truncated
				fclose(outfile);
			}
			if ((outfile = _tfopen(argv[i], _T("wb"))) == NULL )
			{
				if (inpfile)
					fclose(inpfile);
				_ftprintf(stderr, _T("osihfe: Can't open '%s' for output.\r\n"), argv[i]);
				return -1;
			}

		}
		else if (bClean)
		{
			fclose(inpfile);
			_ftprintf(stderr, _T("osihfe: Target filename required for clean operation\r\n"));
			return -1;
		}
	}
	else
	{
		usage();
		return 0;
	}
	/*if (outfile == stdout) //ensure our binary \n's don't convert to \r\n!!!!!
	{
		_setmode(fileno(stdout), _O_BINARY);
	}*/


	if (bBlank)
	{
		if (CreateHFEImage( NULL, dTarget, 2, nBitwidth, &pOSIDisk, &nOSILen, bWrite8As5 ))
		{
			if (bWrite8As5 && pOSIDisk)
			{
				picfileformatheader* pHead = (picfileformatheader*) pOSIDisk;
				if (pHead->bitRate > 250)
					pHead->bitRate = 250;
				pHead->floppyRPM = 0;
			}
			
			size_t nWrote = fwrite(pOSIDisk, 1, nOSILen, outfile);
			if (nWrote != (size_t) nOSILen)
			{
				_ftprintf(stderr, _T("osihfe: Error writing to target file\r\n") );	
			}
			else
			{
				_ftprintf(stderr, _T("osihfe: Wrote file '%s'\r\n"), szOutFile );	
			}
		}


	}
	else
	{

		struct _stat buf;
		int result = _fstat( _fileno(inpfile), &buf );
		size_t nReadBytes = buf.st_size;

		if (result < 0 || nReadBytes < 2000)
		{
			//invalid data
			LogText(_T("Invalid file input size"));
			goto HFEabort;

		}


		pMem = (unsigned char*) ::malloc (2000);
		if (!pMem)
		{
			LogText( _T("Error Can't allocate 2000 bytes"));
		    goto HFEabort;
			
		}
		size_t nRead = fread( pMem, sizeof( char ), 2000, inpfile );

		if (nRead == 2000)
		{
			if (IsHexFile(pMem, 2000))
			{
				bIsHexFile = true;
				free(pMem);
				pMem = NULL;
				int nLen = 0;
				nRead = ReadHexFile(inpfile, &pMem, &nLen);
				nReadBytes = (unsigned int) nLen;
				if (nRead != (size_t)nLen)
				{
					LogText( _T("Exit due to error(s) in HEX file."));
					goto HFEabort;
				}
			}
			if (!memcmp(pMem, "SCP", 3))
			{
				//probable SCP file
				free(pMem);
				pMem = NULL;
			    int nLen = 0;

				if (bIdentify)
					DisplaySCPFileInfo( inpfile);

				//create HFE in-memory from SCP file
				int nRet = CreateHFEFromSCP( inpfile, &pMem, &nLen, nSkipTrack, nSCPRev);
				switch (nRet)
				{
					case 1:
						bIsSCP = true;
						pEnd = pMem + nLen;
						nSkipTrack = 1; //Dont skip tracks on skipped tracks
						break;
					case -1: //memory
						LogText( _T("Error allocating memory"));
						break;
					case -2: //bad format?
						LogText( _T("Error SCP format unexpected (too many heads/tracks etc.)"));
						break;
					case -3: //not scp
						LogText( _T("Error - not SCP file"));
						break;
					case 0:
					case -4: //track load failed
						LogText( _T("Error while loading SCP track"));
						break;
				}
				if (nRet <= 0)
					goto HFEabort;

				if (!bTest && !bClean && !bIdentify && !bDirectory)
				{
					
					if (!outfile)
					{
						//make an output file base on input now that we know what file
						//we are making
						_tcsncpy(szOutFile, szFile, 8191);

						int eLen = _tcslen(szOutFile);
						int j = eLen-1;
						while (	j > 0 && szOutFile[j] != _T('.'))
							j--;
						if (szOutFile[j] == _T('.'))
							szOutFile[j] = 0;
						_tcscat(szOutFile, _T(".hfe"));
					
						//test file exists
						if ( (outfile = _tfopen(szOutFile, _T("r+b"))) != NULL )
						{
							if (!bOverwrite)
							{
								fclose(outfile);
								fclose(inpfile);
								_ftprintf(stderr, _T("osihfe: Target file '%s' exists. Use -o to overwrite\r\n"), szOutFile);
								goto HFEabort;
							}
							//close to reopen truncated
							fclose(outfile);
						}
						if ((outfile = _tfopen(szOutFile, _T("wb"))) == NULL )
						{
							fclose(inpfile);
							_ftprintf(stderr, _T("osihfe: Can't open '%s' for output.\r\n"), szOutFile);
							goto HFEabort;
						}
					}
					TCHAR * pDot = szOutFile + _tcslen(szOutFile) -1;
					while (pDot > szOutFile && *pDot != _T('.'))
						pDot--;
					if (!_tcsicmp(pDot, _T(".hfe")) )
					{
						size_t nWrote = fwrite(pMem, 1, nLen, outfile);
						if (nWrote != (size_t) nLen)
						{
							LogText(_T("Error writing to target file\r\n") );	
						}
						else
						{
							LogText( _T("osihfe: Wrote file '%s'\r\n"), szOutFile);
						}
						fclose(outfile);

						goto HFEabort;
					}
				}



			}
			else
			{
				if (!bIsHexFile)
				{
					free(pMem);
					pMem = NULL;
					fseek(inpfile, 0, SEEK_SET); //reset back to start

					if ( nReadBytes > 10000000)
					{
						//invalid data
						LogText(_T("Invalid HFE file input size"));
						goto HFEabort;

					}
					pMem = (unsigned char*) ::malloc (nReadBytes);
					if (!pMem)
					{
						LogText( _T("Error Can't allocate %d bytes"),nReadBytes);
						goto HFEabort;

					}
					//read data into memory
					nRead = fread( pMem, sizeof( char ), nReadBytes, inpfile );

				}
				pEnd = pMem + nReadBytes;



				if (nRead != nReadBytes)
				{
					if ( ferror(inpfile))
					{
						//error reading from file
						LogText( _T("Error reading from input file\r\n"));
						goto HFEabort;
					}
					else if ( feof(inpfile))
					{
						nReadBytes = nRead;	
					}
				}

				fclose(inpfile); //done reading
				inpfile = NULL;

	

			}


	   }
	   else
	   {
			LogText( _T("Error reading from input file\r\n"));
			goto HFEabort;

	   }

		if (memcmp(pMem, szHeaderSig, 8))	
		{
			//looks like NOT HFE file
			dt = IdentifyDiskType( pMem, nReadBytes);

			if (bIdentify || bDirectory || bTest )
			{
				if (bIdentify && !bTest)
				{
					LogText( _T("File type %s\r\n"), szIdentity[(int) dt]);
					nReturn = dt;
					
				}
				if (bDirectory)
				{
					TCHAR *pText = NULL;
					pText = GetOSIDirectory(pMem, dt);
					
					if (pText)
					{
						
						if (logfile)
						{
							_fputts( pText, logfile);
						}					
						else
						{
							_fputts(pText, stdout);
						}
						free(pText);
						pText = NULL;
					}
					else
					{
						LogText( _T("No directory identified\r\n"));
					}

				}
				if (bTest)
				{
					if (dt == DISK_UNKNOWN)
					{
						_ftprintf(stderr, _T("osihfe: Test - unrecognized input file format\r\n") );
						goto HFEabort;
					}
					
					nReturn = TestOSIImage(pMem, (DISKTYPE) dt, false );
				}

			}
			else
			{

				if (bClean)
				{
					if (CleanOSIImage(pMem, (DISKTYPE) dt,  &pOSIDisk, &nOSILen ))
					{

						if (outfile)
						{
							size_t nWrote = fwrite(pOSIDisk, 1, nOSILen, outfile);
							if (nWrote != (size_t) nOSILen)
							{
								LogText(_T("Error writing to target file\r\n") );	
							}
						}
						else
						{
							_ftprintf(stderr, _T("osihfe: You must specify a target file for Clean\r\n") );
						}


					}
					else
					{
						_ftprintf(stderr, _T("osihfe: Can't clean this disk file.") );
					}
				}
				else
				{
					if (bIsHexFile && !bMakeHFE)
					{
						if (!outfile)
						{
							//make an output file base on input now that we know what file
							//we are making
							int nLen = _tcslen(szFile);
							int j = nLen-1;
							while (	j > 0 && szFile[j] != _T('.'))
								j--;
							if (szFile[j] == _T('.'))
								szFile[j] = 0;						
							switch (dt)
							{
								case OS65D5:
								case OS65D8:
								default:									
									_tcscat(szFile, _T(".65D"));
									break;
								case OS65U8:
									_tcscat(szFile, _T(".65U"));
									break;
							}
							
							if ( (outfile = _tfopen(szFile, _T("r+b"))) != NULL )
							{
								if (!bOverwrite)
								{
									fclose(outfile);
									_ftprintf(stderr, _T("osihfe: Target file '%s' exists. Use -o to overwrite\r\n"), szFile);
									goto HFEabort;
								}
							}
							else if ((outfile = _tfopen(szFile, _T("wb"))) == NULL )
							{
								_ftprintf(stderr, _T("osihfe: Can't open '%s' for output.\r\n"), szFile);
								goto HFEabort;
							}
						}
						size_t nWrote = fwrite(pMem, 1, nReadBytes, outfile);
						if (nWrote != (size_t) nReadBytes)
						{
							_ftprintf(stderr, _T("osihfe: Error writing to target file\r\n") );	
						}
						else
						{
							LogText( _T("osihfe: Wrote file '%s'\r\n"), szFile);
						}
						goto HFEabort;

					}
					if (dt == DISK_UNKNOWN)
					{
						_ftprintf(stderr, _T("osihfe: Unrecognized input file format\r\n"));
						goto HFEabort;
					}
					
					if (CreateHFEImage( pMem, (DISKTYPE) dt, 2, nBitwidth, &pOSIDisk, &nOSILen, bWrite8As5 ))
					{
						if (bWrite8As5 && pOSIDisk)
						{
							picfileformatheader* pHead = (picfileformatheader*) pOSIDisk;
							if (pHead->bitRate > 250)
								pHead->bitRate = 250;
							pHead->floppyRPM = 0;
						}
						
						if (!outfile)
						{
							//make an output file base on input now that we know what file
							//we are making
							_tcsncpy(szOutFile, szFile, 8191);

							int nLen = _tcslen(szOutFile);
							int j = nLen-1;
							while (	j > 0 && szOutFile[j] != _T('.'))
								j--;
							if (szOutFile[j] == _T('.'))
								szOutFile[j] = 0;	
							_tcscat(szOutFile, _T(".HFE"));

							if ( (outfile = _tfopen(szOutFile, _T("r+b"))) != NULL )
							{
								if (!bOverwrite)
								{
									fclose(outfile);
									fclose(inpfile);
									_ftprintf(stderr, _T("osihfe: Target file '%s' exists. Use -o to overwrite\r\n"), szOutFile);
									goto HFEabort;
								}
							}
							else if ((outfile = _tfopen(szOutFile, _T("wb"))) == NULL )
							{
								fclose(inpfile);
								_ftprintf(stderr, _T("osihfe: Can't open '%s' for output.\r\n"), szOutFile);
								goto HFEabort;
							}

								

						}
						
						size_t nWrote = fwrite(pOSIDisk, 1, nOSILen, outfile);
						if (nWrote != (size_t) nOSILen)
						{
							_ftprintf(stderr, _T("osihfe: Error writing to target file\r\n") );	
						}
						else
						{
							LogText( _T("osihfe: Wrote file '%s'\r\n"), szOutFile);
						}
					}

				}
			}


		}
		else
		{
			//found FE header...
			if (bIdentify && !bIsSCP)
				DisplayHFEFileInfo(pMem, nReadBytes);

			//decode HFE src to figure out target type, could read #tracks too
			dt = DetermineOSIDiskTypeFromHFE(pMem, nReadBytes, nSide, nSkipTrack );
			if (dt <= 0)
			{
				_ftprintf(stderr, _T("osihfe: Unable to determine target format from HFE file\r\n") );
				goto HFEabort;	
			}
			if (bIdentify)
			{
				LogText( _T("File type %s\r\n"), szIdentity[(int) dt]);
				nReturn = dt;
				if (!bTest && !bDirectory)
					goto HFEabort;
			}


			if (CreateOSIImage( pMem, nReadBytes, nSide, nSkipTrack, &pOSIDisk, &nOSILen , &dt, bVerbose ))
			{

				if (bTest || bDirectory)
				{
					if (bDirectory)
					{
						
						if (pOSIDisk)
						{
							TCHAR *pText = NULL;
							pText = GetOSIDirectory(pOSIDisk, dt);

							if (pText)
							{
								
								if (logfile)
								{
									_fputts( pText, logfile);
								}					
								else
								{
									_fputts(pText, stdout);
								}
								free(pText);
								pText = NULL;
							}
							else
							{
								LogText( _T("No directory identified\r\n"));
							}

						}
					}
					if (bTest)
						nReturn = TestOSIImage(pOSIDisk, (DISKTYPE) dt, true );
				}
				else
				{
					
					if (!outfile)
					{
						//make an output file base on input now that we know what file
						//we are making
						_tcsncpy(szOutFile, szFile, 8191);
						int nLen = _tcslen(szOutFile);
						int j = nLen-1;
						while (	j > 0 && szOutFile[j] != _T('.'))
							j--;
						if (szFile[j] == _T('.'))
							szFile[j] = 0;
						switch (dt)
						{
							case OS65D5:
							case OS65D8:
							default:									
								_tcscat(szOutFile, _T(".65D"));
								break;
							case OS65U8:
								_tcscat(szOutFile, _T(".65U"));
								break;
						}
						
						if ( (outfile = _tfopen(szOutFile, _T("r+b"))) != NULL )
						{
							if (!bOverwrite)
							{
								_ftprintf(stderr, _T("osihfe: Target file '%s' exists. Use -o to overwrite\r\n"), szOutFile);
								goto HFEabort;
							}
							//close to reopen truncated
							fclose(outfile);
						}
						if ((outfile = _tfopen(szOutFile, _T("wb"))) == NULL )
						{
							_ftprintf(stderr, _T("osihfe: Can't open '%s' for output.\r\n"), szOutFile);
							goto HFEabort;
						}
					}

					if (bClean)
					{	
						unsigned char *pCleaned = NULL;
						if (CleanOSIImage(pMem, (DISKTYPE) dt,  &pCleaned, &nOSILen ))
						{
							if (pOSIDisk)
							{
								free(pOSIDisk);

							}
							pOSIDisk = pCleaned;	
						}
					}


					size_t nWrote = fwrite(pOSIDisk, 1, nOSILen, outfile);
					if (nWrote != (size_t) nOSILen)
					{
						_ftprintf(stderr, _T("osihfe: Error writing to target file\r\n") );	
					}
					else
					{
						LogText( _T("osihfe: Wrote file '%s'\r\n"), szOutFile);
					}
				}
			}
		}
	}

HFEabort:

	if (pOSIDisk)
	{
		free(pOSIDisk);
		pOSIDisk = NULL;
	}

	if (pMem)
	{
		free(pMem);
		pMem = NULL;
	}
	if (inpfile)
		fclose(inpfile); //done reading
	
	if (outfile)
		fclose(outfile);
	if (logfile != NULL && logfile != stderr)
		fclose(logfile);
	
	return nReturn;
}
static int logpos = 0;
void LogData(unsigned char c)
{
	TCHAR hex[17] = _T("0123456789ABCDEF");
	TCHAR out[4] = {0};
    if (logfile)
	{
		
		if ((logpos%16) == 0)
		{
			_fputts( _T("\r\n"), logfile);
			logpos = 0;
		}
		logpos++;

		out[0] = hex[c >> 4];
		out[1] = hex[c &0x0F];
		out[2] = 0x20;
		out[3] = 0;
		_fputts( out, logfile);
	}
}
void LogText(const TCHAR * szFormat, ...)
{
	static TCHAR buf[2048];
    va_list vargs;

    if (logfile)
	{
		if (szFormat && *szFormat) 
		{
			va_start( vargs, szFormat );
			_vsntprintf( buf, 2047, szFormat, vargs );
			va_end( vargs );  
        
			_fputts( buf, logfile);
			logpos = 0;
		}
	}
}

int IsHexLine(unsigned char *pMem, int nLen)
{
	int i = 0;
	unsigned char c1, c2, c3;
	int nHex = 0;
	while (i < nLen)
	{
		c1 = *(pMem+i);
		c2 = i < nLen-1 ? *(pMem+i+1) : (unsigned char )0;
		c3 = i < nLen-2 ? *(pMem+i+2) : (unsigned char )0;

		if (c1 == 0x20 || c1 == 0x09)
		{
			i++;
			continue;
		}
		if (c1 == 0x0a || c1 == 0x0d)
		{
			return nHex;
		}
		//handle data corruption
		//22 00 0D A2 C3 86 00 A2 28 86 01 A2 22 86 02 A0
		//04 8C 01 F7 A2 FF 8E00 F7 E8 8E 01 F7 CA 8E 

		if (isxdigit(c1) && isxdigit(c2) && ( c3 == 0x20 || c3 == 0x0D || c3 == 0x0A || c3 == 0) )
		{
			nHex++;
			i+=3;
			if (c3 == 0x0D || c3 == 0x0A || c3 == 0)
				return nHex;
		}
		else
		{
			return -nHex;
		}
		 
	}
	return nHex;
}
int DecodeHexLine(unsigned char *pMem, int nLen, unsigned char *pDest)
{
	int i = 0;
	unsigned char c1, c2;
	int nHex = 0;
	while (i < nLen)
	{
		c1 = *(pMem+i);
		c2 = i < nLen-1 ? *(pMem+i+1) : (unsigned char )0;

		if (c1 == 0x20 || c1 == 0x09)
		{
			i++;
			continue;
		}
		if (c1 == 0x0a || c1 == 0x0d)
		{
			return nHex;
		}
		if (isxdigit(c1) && isxdigit(c2))
		{
			if (c1 >= 'A')
				c1 -=7;
			if (c2 >= 'A')
				c2 -=7;
			c1 &= 0x0F;
			c2 &= 0x0F;

			*pDest++ = (unsigned char) ( (c1 << 4) | c2 );
			i+=2;
			nHex++;
		}
		else
		{
			return 0;
		}
		 
	}
	return nHex;

}

bool IsHexFile(unsigned char * pMem, int nLen)
{
	int n = 0, nPos = 0, nCount = 0, nLines = 0;

	if (!pMem || nLen < 4)
		return false;
	
	while ( pMem &&  n < nLen-1 )
	{
		unsigned char c1 = pMem[n];
		if (c1 == '\r' || c1 == '\n')
		{
			if (nPos)
			{
				if (nPos >= 44 && IsHexLine(pMem+n-nPos, nPos) >= 16)
				{
					nCount++;	
				}
			}
			nPos = 0;
			if (pMem[n+1] != '\n')
				nLines++;
		}
		else
		{
			nPos++;
		}
		n++;
	}

	return nCount > 0 ;

}


int ReadHexFile( FILE* inFile, unsigned char **ppMem, int* pnLen)
{
	if (!ppMem || !pnLen || !inFile)
		return -1;

	*ppMem = NULL;
	*pnLen = 0;

	fseek(inFile, 0, SEEK_END);
	long lSize = ftell(inFile);

	if (lSize > 2000000)
	{
		_fputts( _T("Unexpected file size for hex conversion (too big).\r\n"), stderr);
		
		//file too big
		return -1;
	}

	unsigned char * pBuf = (unsigned char *) malloc(lSize+1);
	if (!pBuf)
	{
		//error alloc ram
		return -1;
	}
	memset(pBuf, 0, lSize+1);

	fseek(inFile, 0, SEEK_SET); //reset back to start
	long nRead = fread( pBuf, sizeof( char ), lSize, inFile );

	if (nRead != lSize)
	{
		//error read
		free(pBuf);
		_fputts( _T("Error reading from HEX file\r\n"), stderr);
		return -1;
	}
	*ppMem = (unsigned char *) malloc(lSize/2+100);

	if (!*ppMem)
	{
		//error alloc ram
		free(pBuf);
		_ftprintf(stderr, _T("osihfe: Error allocating %d bytes"),(int) (lSize/2+100) );
		return -1;
	}
	memset(*ppMem, 0, lSize/2+100);

	unsigned char * pDest = *ppMem;

	long n = 0;
	int nLine = 1;
	int nTrack = -1;
	int nPos = 0, nWrote =0, nTrackLen = 0;
	bool bDecoding = false;
	bool bError = false;
	int nTrackSize = 0;
	 
	while ( n < lSize )
	{

		unsigned char c1 = pBuf[n];
		if (c1 == '\r' || c1 == '\n')
		{
			if (n && pBuf[n-1] != '\r')
				nLine++;
			if (nPos)
			{
				if ( !memcmp(pBuf+n-nPos, ";READING TRACK ", 15))
				{
					//maybe pad tracks to correct size when we know where they start?
					if (nTrackLen && nTrack >= 0 && nTrackSize == 0)
					{
						if (nTrackLen > TRKSIZ5 + 32)
							nTrackSize = TRKSIZ8; 
						else
							nTrackSize = TRKSIZ5;	
					}
					else if (nTrackLen < nTrackSize)
					{
						pDest += (nTrackSize - nTrackLen);
					}
					
					nTrack = atoi((const char *) pBuf+n-nPos+15);
					nTrackLen = 0;
				}
				else if (nPos >= 44 && IsHexLine(pBuf+n-nPos, nPos) >= 16)
				{
					int nVal = DecodeHexLine(pBuf+n-nPos, nPos, pDest);
					pDest += nVal;
					nWrote += nVal;
					nTrackLen += nVal;
					bDecoding = true;
				}
				else
				{
					if ( bDecoding && *(pBuf+n-nPos) == ';')
						bDecoding = false;

					if (bDecoding)
					{
						_ftprintf(stderr, _T("osihfe: Unexpected data in HEX stream @ line %d"),nLine );
						if (nTrack >= 0)
							_ftprintf(stderr, _T(", (Track %d)"), nTrack);
						_ftprintf(stderr, _T("\r\n"));	
						bDecoding = false;
						bError = true;
					}
				}
			}
			nPos = 0;
		}
		else
		{
			nPos++;
		}
		n++;
	}

	free(pBuf);

	*pnLen = nWrote;

	return bError ? -nWrote: nWrote;

}
