#include "scp.h"
#include <stdio.h>
#include <stdlib.h>

#define SCPBUFF 8192
extern bool bVerbose;
extern FILE* logfile;

//HFE file uses a constant bitrate,  SCP data stores length of time of flux transitions in sample rate time unit
//so 500,000 HFE bitrate each nibble is 1 flux at 2us (2000ns)
//9E, 9E, A0, 9F, 9B, 51, 50, 4E, 57, 98, 57	  158 * 25ns = 3950
//1100 1100 1100 1100 1100 1111 1111 1111 1111 1100 1111
//
//0010 0010 0010 0010 0101 0101 0101 0101 0101 0100 0101

// HFE 8" bitrate 500, 5.25" bitrate 250;
// resolution in nanoseconds
// bitwidth is size of source bits, resolution is source sample witdh in ns, nOffset is position of TDH
// revs is number of revolutions sampled (usually 3) nSCPRev is revolution we are processing
// ppFlux is pointer to decoded flux in HFE format, pnLen is the size of decoded flux

bool LoadSCPTrack( FILE* inFile, unsigned char bitwidth, unsigned char resolution, int nOffset, int revs, int nSCPRev, unsigned char **ppFlux, int *pnLen )
{
	scptrackdataheader* pthdr = NULL;
	unsigned char *pMem = NULL;
	unsigned char *pHFETrk = NULL;
	int nLen = 0;
	if (pnLen)
		*pnLen = 0;
	if (ppFlux)
		*ppFlux = NULL;

	typedef struct hdrarry_
	{
		unsigned long  indextime;	 //0x04-0x07 LE duration of 1 revolution in samples ex: (6666667) * 25ns = 166666675ns = 166.666675ms,  
									 //          60000ms per min / 166.666675 = 360 RPM
		unsigned long  tracklen;		 //0x08-0x0B LE length of track in bitcells 16bits = 2 & bytelen
		unsigned long  trackoffset;	 //0x0C-0x0F LE offset from start of TDH to flux data

	} hdrarry;

	bool bWrote = false;

	if (nOffset == 0 || nOffset < sizeof(scpfileformatheader))
		return NULL; //illegal or not present TDH

	if ( !fseek(inFile, nOffset, SEEK_SET) )
	{
		if (revs <= 0)
			revs = 1;
		if (nSCPRev < 0 || nSCPRev > revs)
			nSCPRev = revs;

		unsigned int thdrsize = sizeof(scptrackdataheader) + (revs-1)* sizeof( hdrarry);
		pthdr = ( scptrackdataheader*) malloc(thdrsize);
		
		
		if (pthdr && (fread(pthdr, 1, thdrsize,inFile) == thdrsize) )
		{
			if (!memcmp( pthdr->HEADERSIGNATURE, "TRK", 3) && pthdr->tracklen < 0x100000 )
			{
				unsigned long trkoffset = pthdr->trackoffset;
				unsigned long trklength = pthdr->tracklen;
				unsigned long trkindextime = pthdr->indextime; 
				if (nSCPRev > 0)
				{
					hdrarry* pHdr = (hdrarry*)(((unsigned char *) pthdr)+4);
					trkoffset = pHdr[nSCPRev-1].trackoffset;
					trklength = pHdr[nSCPRev-1].tracklen;
					trkindextime = pHdr[nSCPRev-1].indextime;
				}
				
				
				//nBitwidth usually 16, but could be 8 or 32 or partial bit len???
				int nCellPad = (bitwidth+7)/8; //pad to bytelen

				pMem = (unsigned char *) malloc( trklength * nCellPad);
				if (!pMem)
					return false;
				fseek(inFile, nOffset+trkoffset, SEEK_SET);
				//fseek(inFile, trkoffset, SEEK_CUR);

				int samplerate = resolution == 0 ? 25 : 25 + 25*resolution;
				nLen = (DISK8TARGETLEN + 511) / 512 * 512; 
				int nTargetLen = DISK8TARGETLEN;
				int indxms = (trkindextime/1000) * samplerate / 1000;
				int fluxrate = 2000; //1
				if (indxms > 180)  //8" drive = 360RPM (166 ms) *500K =83000 .166/83000=2us    5.25=300 RPM = 200 ms * 250K = .200/50000 = 4us
				{
					nLen = (DISK5TARGETLEN + 511) / 512 * 512; 
					nTargetLen = DISK5TARGETLEN;
					fluxrate = 4000;
				}

				
				//8" flux cell = 2us, 2000ns
				//5.25" flux cell = 4us, 4000ns
				unsigned int totaltracktime = 0;
				if (fread(pMem, 1, trklength * nCellPad ,inFile))
				{
					//The words used for the flux length in the actual flux data are in big endian format
					
					pHFETrk = (unsigned char *) malloc(nLen );
					if (pHFETrk)
					{
						unsigned char * pData = pMem;
						unsigned int i;
						unsigned short  t = 0;
						int time = 0;
						unsigned char *p = pHFETrk;
						unsigned char *pEnd = pHFETrk+ nTargetLen;
						//int celltimepersample = (fluxrate/samplerate);
						//int halfcelltime = 	celltimepersample/2;

						memset(pHFETrk, 0, nLen);
						int cellidx = 0;
						for (i = 0; i < trklength && totaltracktime < trkindextime; i++)
						{
							t = (unsigned short)( ((*pData) << 8) + *(pData+1)); pData +=2;
							totaltracktime += t;
							
							if (t == 0)
							{
								//no flux transition for an extended period
								time += 2^bitwidth;
								totaltracktime += 2^bitwidth;	//usually 65536

							}
							else
							{
								/*
								//Alternate way of decoding based on absolute time offsets
								//(note: would need to account for flux rate diffences between ideal and actual due to RPM variances)
								int pos = (totaltracktime+halfcelltime)/celltimepersample;
								int cell = pos / 4;
								if (cell > DISK8TARGETLEN)
									break;
								int bit = pos & 0x03;
								switch (bit)
								{
									case 0:
										pHFETrk[cell] |=  0x01;
										break;
									case 1:
										pHFETrk[cell] |= 0x04;
										break;
									case 2:
										pHFETrk[cell] |= 0x10;
										break;
									case 3:
										pHFETrk[cell] |= 0x40;
										break;
								}


								
								*/
								//sloppy way of decoding based on sample increments
								//even at different rates of rotation the number of flux transitions should be constant
								//the converted image has ideal flux rate, the source could vary
								int span = ((t+time) * samplerate+999); //flux len in ns + some rounding
								 
								int advance = span/fluxrate;
								cellidx += advance;
								p+= (cellidx/4);
								cellidx &= 0x03;
								if (p>= pEnd)
									break;
								bWrote = true;
								switch (cellidx)
								{
									case 0:
										*p |=  0x01;
										break;
									case 1:
										*p |= 0x04;
										break;
									case 2:
										*p |= 0x10;
										break;
									case 3:
										*p |= 0x40;
										break;
								}
								
								 time = 0;


							}
						
						}

					}
					else
					{
						//mem failure
					}	

				}
				else
				{
					//read failure
				}
			
#if 0
				//debug - test decoded stream
				if (bWrote)
				{
					FMstate fm;
					FMdecodeInitialize(fm);
					fm.eSformat = s8E1;

					int nRead, nWrote, pos = 0;
					FMdecodestream(pHFETrk, nLen, 2, pMem, NULL, thdr.tracklen * nCellPad,fm, &nRead, &nWrote, pos, bVerbose);
				}
#endif
				if (pMem)
					free(pMem);
				pMem = NULL;

			}
			else
			{
				//TRK Header not found
			}

		}
		if (pthdr)
		{
			free(pthdr);
			pthdr = NULL;
		}
	}

	if (ppFlux)
	{
		*ppFlux = pHFETrk;
	}
	else
	{
		free(pHFETrk);
		pHFETrk = NULL;
	}
	if (pnLen)
		*pnLen = nLen;

	return bWrote;
	 
}


// Display information abour SCP file from header
// we can calculate & display rotational speed and flux rate
bool DisplaySCPFileInfo( FILE* inFile)
{
	bool bOK = false;
	bool bReadError = false;
	scptrackdataheader thdr = {0};
	const TCHAR *sTPI = _T("48TPI");
	const TCHAR *sRPM = _T("300RPM");
	TCHAR buf[2048];

	if (!inFile)
		return false;
	
	unsigned char * pBuf = (unsigned char *) malloc(SCPBUFF);
	if (!pBuf)
		return false;


	//we only convert Full Track single or double sided SCP images
	fseek(inFile, 0, SEEK_SET); //reset back to start

	if (fread(pBuf, 1, SCPBUFF,inFile) == SCPBUFF)
	{
		scpfileformatheader *pSCP = (scpfileformatheader*) pBuf;

		if (!memcmp( pSCP->HEADERSIGNATURE, "SCP", 3) )
		{
			int curtrk = 0;
			int cellwidth = pSCP->bitcellwidth == 0 ? 16 : pSCP->bitcellwidth;

			if ((pSCP->scpflag & 0x01) != 0 ) sTPI = _T("96TPI");
			if ((pSCP->scpflag & 0x02) != 0 ) sRPM = _T("360RPM");

			int nOffset = pSCP->TDH[curtrk];
			
			if (nOffset > sizeof(scpfileformatheader))
			{
				if ( !fseek(inFile, nOffset, SEEK_SET) )
				{
					if (fread(&thdr, 1, sizeof(scptrackdataheader),inFile) == sizeof(scptrackdataheader))
					{
						if (!memcmp( thdr.HEADERSIGNATURE, "TRK", 3) && thdr.tracklen < 0x100000 )
						{

							bOK = true;
						
							int samplerate = (pSCP->resolution == 0) ? 25 : 25 + 25*pSCP->resolution;
							int indxms = (thdr.indextime/1000) * samplerate / 1000;
							int fluxrate = 2000;
							if (indxms > 180)  //8" drive = 360RPM (166 ms) *500K =83000 .166/83000=2us    5.25=300 RPM = 200 ms * 250K = 50000 flux .200/50000 = 4us 4000ns
							{
								fluxrate = 4000;
							}

							//nBitwidth usually 16, but could be 8 or 32 or partial bit len???
							int nCellPad = (cellwidth+7)/8; //pad to bytelen

							unsigned char *pMem = (unsigned char *) malloc( thdr.tracklen * nCellPad);
							if (pMem)
							{

								fseek(inFile, thdr.trackoffset, SEEK_CUR);

								//read ist track into memory
								unsigned int totaltracktime = 0;
								if (fread(pMem, 1, thdr.tracklen * nCellPad ,inFile))
								{

									int  t;
									unsigned int i;
									unsigned char * pData = pMem;
									int shortsum = fluxrate/samplerate;
									int longsum = shortsum*2;

									int shortcount = 1;
									int longcount = 1;
									int samplecount = 1;
									//calc average short flux time & long flux time
									for (i = 0; i < thdr.tracklen && totaltracktime < thdr.indextime; i++)
									{
										t = (unsigned short)( ((*pData) << 8) + *(pData+1)); pData +=2;
										totaltracktime += t;
										
										if (t == 0)	//0 indicates overflow at sample resolution
										{
											//no flux transition for an extended period
											totaltracktime += 2^cellwidth;	//usually 65536
										}
										else
										{
											//which "slot" should this sample go into?
											samplecount++;
											if (t > (shortsum/shortcount * 3 / 2)  )  //fluxrate * 1.5 = long flux span
											{
												longsum += t;
												longcount++;
											}
											else
											{
												shortsum +=t;
												shortcount++;
											}
	
										}
									}

									int x = shortsum/shortcount*samplerate;
									if (!x) x=1;

#if (_MSC_VER >= 1400) && (defined(UNICODE))
									swprintf(buf, _T("SCPFile: tracks %d to %d, %d sides, %d revs, bitrate:%0.2fKfps, %s, %s (header) %dRPM actual\r\n"),(int)pSCP->start_track, (int) pSCP->end_track,
										(int) (pSCP->heads == 2) ? 2 : 1, (int)pSCP->revs, (double)1000000/x, sTPI, sRPM, indxms ? 60000/indxms : 0);
#else							
									_stprintf(buf, _T("SCPFile: tracks %d to %d, %d sides, %d revs, bitrate:%0.2fKfps, %s, %s (header) %dRPM actual\r\n"),(int)pSCP->start_track, (int) pSCP->end_track,
										(int) (pSCP->heads == 2) ? 2 : 1, (int)pSCP->revs, (double)1000000/x, sTPI, sRPM , indxms ? 60000/indxms : 0);
#endif
									if (logfile)
									{
										_fputts( buf, logfile);
									}					
									else
									{
										_fputts(buf, stdout);
									}
								}
								else
								{
									//read failure
									bReadError = true;
								}
								free (pMem);
							}
							else
							{
								//mem failure
								LogText( _T("Error allocating memory for SCP file"));
							}
						}
						else
						{
							//invalid trk data
							bOK = false;
						}
					}
					else
					{
						bReadError = true;
					}
				}
				else
				{
					bReadError = true;	
				}
			
			}
			else
			{
				//invalid header data
				bOK = false;
			}

		}
		else
		{
			//invalid file format
			bOK = false;
		}
	}
	if (bReadError)
	{
		LogText( _T("Error reading SCP file"));
	}

	free(pBuf);
	return bOK;

}
/*
unsigned char HEADERSIGNATURE[3]; // "SCP"
unsigned char version;			// Revision 0.0
unsigned char disktype;			// Disk Class & Manufacturer Class. 0x80 = other/360K or 0x84 other/720K?
unsigned char revs;				// revolutions per track (number of track samples) (3+)
unsigned char start_track;		// starting track
unsigned char end_track;		// ending track (2x actual track if using half tracks) max 168
unsigned char scpflag;			// flags - see below
unsigned char bitcellwidth;		// 0=16bits, else value = # of bits for sample
unsigned char heads;			// 0=2 sided, 1= side 0 (bottom), 2= side 1 (top)
unsigned char resolution;		// 0=25ns else (25 + 25*resolution) ns
unsigned int  checksum;	//little endian 32bit checksum offset 0x10 to end of file except if scpflag bit 4 set
long TDH[1];  //table of offsets to track data header * number of tracks
              //If no flux data for a track is present, then the entry will contain a longword of zeros (0x00000000)
//scpflag:
//bit 0 INDEX, when set indicates track data starts at index detection
//bit 1 TPI,  clear = 48TPI, set = 96TPI
//bit 2 RPM,  clear = 300RPM, set=360 RPM
//bit 3 TYPE, clear for preservation, set for normalization
//bit 4 MODE, clear = read-only, set = R/W
//bit 5 FOOTER, clear if not extension footer, set if extension footer present
//bit 6 EXTENDED MODE, clear for floppy drives, set for other media


} scpfileformatheader;

*/
//Create in-memory conversion of SCP to HFE
//
// returns:
//   1 OK
//   0 cant read scp
//	-1 memory failure
//  -2 unrecognized format or invalid number of tracks/sides
//  -3 not SCP file
//  -4 track conversion failed
//  

int CreateHFEFromSCP( FILE* inFile, unsigned char **ppHFE, int* pnLen, int nSkip /*=1*/, int nSCPRev /*=1*/)
{
	int nReturn = 0;

	//lets start with 8K
	if (!ppHFE || !pnLen || !inFile)
		return -1;
	*ppHFE = NULL;
	*pnLen = 0;

	DISKTYPE dt = OS65D8;
	unsigned char * pBuf = (unsigned char *) malloc(SCPBUFF);

	if (!pBuf)
		return -1;
	//we only convert Full Track single or double sided SCP images
	fseek(inFile, 0, SEEK_SET); //reset back to start

	if (fread(pBuf, 1, SCPBUFF,inFile) == SCPBUFF)
	{
		scpfileformatheader *pSCP = (scpfileformatheader*) pBuf;

		if (!memcmp( pSCP->HEADERSIGNATURE, "SCP", 3) )
		{
			//validate header some, no invalid track ranges
			if ( pSCP->end_track < pSCP->start_track || pSCP->end_track > 168 || pSCP->heads > 2) 
			{
				nReturn = -2;
				goto scpexit;
			}

			if (nSCPRev <= 0)
				nSCPRev = pSCP->revs;
			if (nSCPRev > pSCP->revs)
			{
				LogText( _T("Error specified SCP revolution %d does not exist\r\n"),nSCPRev);
				nReturn = -4;
				goto scpexit;
			} 

			if (pSCP->disktype == 0x80) dt = OS65D5;

			//some SCP files lie, claim single sided but contain 2 side of data.
			bool b2Sides = pSCP->heads == 0 ? true : false;
			int nStartTrack = 0;
			int nEndTrack = pSCP->end_track;
			if (nSkip>1)
				nEndTrack = (pSCP->end_track)/nSkip; 
			else 
			{
				if (pSCP->end_track > 81 || b2Sides ) //could be 40 track or incomplete image
				{
					b2Sides = true;
					nEndTrack = pSCP->end_track / 2;
				} 
				else if ((pSCP->scpflag & 0x02) && (pSCP->end_track ==78 || pSCP->end_track == 79) )	  //96TPI image?
				{
					//96TPI 40 track disk
					if (pSCP->TDH[1] == 0)
					{
						nSkip = 2; //pretend to be 2 sided disk to skip every other track
						nEndTrack = pSCP->end_track / 2;
						nStartTrack = pSCP->start_track / 2;
					}
				}
				if (b2Sides)
					nStartTrack = pSCP->start_track / 2;
			}


			//if SCP is stored in half-tracks, 
			int nTracks = nEndTrack - nStartTrack + 1;

			
			int nSides = 2;
			
			int nDiskTrackLen = DISK8TARGETLEN;
			
			//need to allocate HFE image
			
			unsigned int dwHeader = 512; //sizeof(picfileformatheader); 
			unsigned int  dwTrackTable =  (sizeof(pictrack) * nTracks);
			dwTrackTable = ((dwTrackTable + 511) / 512) * 512; //padded table size
			//track data is padded
			nDiskTrackLen = (nDiskTrackLen + 511) / 512 * 512;

			unsigned int nTrackTotalSize = nDiskTrackLen * nTracks * nSides;

			//let make an HFE image;
			 unsigned char *pHFE = (unsigned char *) malloc( dwHeader + dwTrackTable + nTrackTotalSize);
			if (pHFE)
			{
				memset(pHFE, 0, dwHeader + dwTrackTable + nTrackTotalSize);
				memset(pHFE, 0xFF, dwHeader+dwTrackTable);


				picfileformatheader* pHead = (picfileformatheader*) pHFE;
				memcpy(pHead->HEADERSIGNATURE, szHeaderSig, sizeof(pHead->HEADERSIGNATURE) );
				pHead->formatrevision = 0;
				pHead->number_of_track = (unsigned char)nTracks;
				pHead->number_of_side = (unsigned char)nSides;
				pHead->track_encoding = 0xFF;
				pHead->bitRate = (unsigned short)(dt == OS65D5 ? 250 : 500);
				pHead->floppyRPM = (unsigned short)(dt == OS65D5 ? 300 : 360);
				pHead->floppyinterfacemode = GENERIC_SHUGGART_DD_FLOPPYMODE;
				pHead->dnu = 1; //(?)
				pHead->track_list_offset = 1; //(=512)
				pHead->write_allowed = 0xFF;
				pHead->single_step = 0xFF;
				pHead->track0s0_altencoding = 0xFF;
				pHead->track0s0_encoding = 0xFF;
				pHead->track0s1_altencoding = 0xFF;
				pHead->track0s1_encoding = 0xFF;

				pictrack* TrkLst = (pictrack*)(pHFE + 0x200);

				unsigned char* pFluxTrack = NULL;

				int curtrk = 0;
				int cellwidth = pSCP->bitcellwidth == 0 ? 16 : pSCP->bitcellwidth;
				int rev = 2;

				if (pSCP->revs > 1)
					rev = pSCP->revs -1;

				unsigned char *pHFETracks = (unsigned char *) (pHFE + dwTrackTable + dwHeader);
				nReturn = 1;
				unsigned int nCurrentTotalSize = 0;
				//make a list of track lengths
				for (int i = 0; i < nTracks; i++)
				{
					int nLen = 0;
					if (i >= nStartTrack) 
					{
											
						if (!LoadSCPTrack(inFile, (unsigned char) cellwidth, (unsigned char)  pSCP->resolution, pSCP->TDH[curtrk], pSCP->revs, nSCPRev, &pFluxTrack, &nLen ))
						{
							//track conversion failed
							nReturn = -4;
							if (bVerbose)
								LogText( _T("Error converting SCP track %d\r\n"),i);
						}

						
					}
					if (nSkip > 1)
						curtrk += nSkip;
					else
						curtrk += b2Sides ? 2 : 1;
					

					//now interleave blocks to create double sided data
					TrkLst[i].offset = (unsigned short) ((pHFETracks - pHFE) / 512);
					TrkLst[i].track_len = (unsigned short) (nLen*nSides);

					int nPadded = (nLen*nSides + 511) / 512 * 512;
					nCurrentTotalSize += nPadded;

					int nBlocks = (nLen + 255) / 256;
					int remainder = nLen;
					if (nSides == 2) //interleave blocks
					{
						unsigned char *p = pFluxTrack;
						unsigned char *pOut = pHFETracks;
						for (int j = 0; j < nBlocks; j++)
						{
							if (p)
							{
								memcpy(pOut, p, remainder >= 256 ? 256 : remainder);	//copy side 1 block
								p += 256;
								remainder -=256;
								if (remainder < 0)
								{
									remainder = 0;
									p = NULL;
								}
							}
							else 
								memset(pOut, 0x11, 256); //fill with FM 0's (blank)
							pOut += 256;
							memset(pOut, 0x11, 256); //fill side 2 with FM 0's (blank)
							pOut += 256; 	
							
						}

					}
					else   //single data block
					{
						if (pFluxTrack)
							memcpy(pHFETracks, pFluxTrack, nLen);
						else
						    memset(pHFETracks, 0x11, nLen); //fill side 2 with FM 0's (blank)

					}
					if (pFluxTrack)
					{
						free(pFluxTrack);
						pFluxTrack = NULL;
					}
					
					pHFETracks += nPadded;
				} //for

				if (nCurrentTotalSize > nTrackTotalSize)
				{
					nReturn = -99;	//sanity check
					goto scpexit;
				}
				
				if (pnLen )
					*pnLen = dwHeader + dwTrackTable + nCurrentTotalSize;

				if (ppHFE)
				{
					*ppHFE = pHFE;
				}
				else
				{
					free(pHFE);
					pHFE = NULL;
				}
				return nReturn;
			}
			else
			{
				return -1; //can't alloc memory
			}

			

		}	//if SCP data
		else
		{
			return -3; //not scp
		}


	} //if READ
scpexit:
	if (pBuf)
	{
		free(pBuf);
		pBuf = NULL;
	}
	return nReturn;


}